mod gobject;
mod relm;
mod zipline;

use std::{borrow::Cow, path::PathBuf, time::Duration};

use color_eyre::eyre::{OptionExt, Result, bail};
use gobject::GtkZiplineFolder;
use relm::{Dialog, DialogInput};
use relm4::{
    Component, ComponentController, Controller, RelmApp,
    adw::{self, prelude::*},
    gtk::{self, gio},
    prelude::{AsyncComponent, AsyncComponentParts},
    tokio,
};
use tracing::Level;
use tracing_subscriber::EnvFilter;
use urlencoding::Encoded;
use zipline::ZiplineFolder;

#[derive(Debug, Default, Clone, serde::Deserialize, serde::Serialize)]
struct Config {
    zipline_url: String,
    zipline_token: String,
}

impl Config {
    fn fixed_url(&self) -> String {
        let mut url = self.zipline_url.clone();
        if !url.ends_with("/") {
            url += "/";
        }
        if url.ends_with("api/") {
            url.truncate(url.len() - 3);
        }

        url
    }
}

#[derive(Debug)]
enum Message {
    OpenFilePicker,
    SetFolder(ZiplineFolder),
    LockAndStart,
    StartTheProcess,
    Nothing,
}

struct Tyrolienne {
    config: Config,
    locked: bool,
    video_path: Option<PathBuf>,
    folder: Option<ZiplineFolder>,
    dialog: Controller<Dialog>,
}

impl Tyrolienne {
    fn display_video_path(&self) -> Cow<'_, str> {
        self.video_path
            .as_ref()
            .map(|p| p.to_string_lossy())
            .unwrap_or(Cow::Borrowed("None"))
    }
}

#[relm4::component(async)]
impl AsyncComponent for Tyrolienne {
    type Input = Message;
    type Output = ();
    type CommandOutput = ();
    type Init = Config;

    view! {
        adw::ApplicationWindow {
            set_title: Some("Tyrolienne"),
            set_default_width: 500,

            gtk::Box {
                set_orientation: gtk::Orientation::Vertical,
                set_spacing: 0,

                adw::HeaderBar {},

                gtk::Box {
                    set_orientation: gtk::Orientation::Vertical,
                    set_spacing: 32,
                    set_margin_top: 32,
                    set_margin_bottom: 32,
                    set_margin_start: 32,
                    set_margin_end: 32,

                    gtk::ListBox {
                        set_selection_mode: gtk::SelectionMode::None,
                        set_css_classes: &["boxed-list"],

                        adw::ActionRow {
                            set_activatable: true,
                            set_css_classes: &["property"],
                            set_title: "Video file",
                            #[watch]
                            set_subtitle: &model.display_video_path(),
                            connect_activated => Message::OpenFilePicker,
                        },

                        adw::ComboRow {
                            set_title: "Folder",
                            set_model: Some(&folder_store),
                            set_expression: Some(&folder_expression),
                            connect_activated[sender] => move |r| {
                                if let Some(item) = r
                                    .selected_item()
                                    .and_then(|i| i.downcast::<GtkZiplineFolder>().ok())
                                {
                                    sender.input(Message::SetFolder(item.as_folder()));
                                }
                            },
                        }
                    },

                    gtk::Button {
                        #[watch]
                        set_label: if model.locked { "Uploading..." } else { "Send" },
                        #[watch]
                        set_sensitive: !model.locked && model.video_path.is_some(),
                        connect_clicked => Message::LockAndStart,
                    }
                }
            }
        }
    }

    async fn init(
        config: Self::Init,
        root: Self::Root,
        sender: relm4::AsyncComponentSender<Self>,
    ) -> AsyncComponentParts<Self> {
        let model = Tyrolienne {
            config,
            locked: false,
            video_path: None,
            folder: None,
            dialog: Dialog::builder()
                .launch(root.clone())
                .forward(sender.input_sender(), |_| Message::Nothing),
        };

        // TODO consider using the "loading" (?) mechanism from relm4
        // https://relm4.org/book/stable/threads_and_async/async.html
        let folders = match zipline::get_folders(&model.config).await {
            Ok(v) => v,
            Err(e) => {
                model.dialog.emit(DialogInput::Show {
                    heading: "Could not fetch folders".into(),
                    body: e.to_string(),
                });
                Vec::new()
            }
        };

        let mut gtk_folders = folders
            .iter()
            .map(GtkZiplineFolder::from_folder)
            .collect::<Vec<_>>();
        gtk_folders.insert(0, GtkZiplineFolder::none());

        let folder_store = gio::ListStore::new::<GtkZiplineFolder>();
        folder_store.extend_from_slice(&gtk_folders);

        let folder_expression = gtk::PropertyExpression::new(
            GtkZiplineFolder::r#type(),
            None::<&gtk::Expression>,
            "name",
        );

        let widgets = view_output!();
        AsyncComponentParts { model, widgets }
    }

    async fn update(
        &mut self,
        message: Self::Input,
        sender: relm4::AsyncComponentSender<Self>,
        _root: &Self::Root,
    ) {
        match message {
            Message::Nothing => {}
            Message::SetFolder(folder) => {
                self.folder = (folder.id != GtkZiplineFolder::NONE_ID).then_some(folder)
            }
            Message::OpenFilePicker => {
                let file = rfd::AsyncFileDialog::new()
                    .add_filter("Video file", &["mp4", "mkv", "webm"])
                    .pick_file()
                    .await;

                if let Some(file) = file {
                    self.video_path = Some(file.path().to_owned());
                }
            }
            Message::LockAndStart => {
                // a little bit convoluted, but i don't really know how to force a view update
                // before starting the "long" process
                self.locked = true;
                sender.input(Message::StartTheProcess);
            }
            Message::StartTheProcess => {
                match the_process(self).await {
                    Ok(url) => tracing::info!("{url}"),
                    Err(e) => self.dialog.emit(DialogInput::Show {
                        heading: "An error occurred".into(),
                        body: e.to_string(),
                    }),
                }
                self.locked = false;
            }
        }
    }
}

fn main() -> Result<()> {
    tracing_subscriber::fmt()
        .with_env_filter(EnvFilter::from_default_env().add_directive(Level::INFO.into()))
        .init();

    color_eyre::install()?;

    // TODO: show dialog in case these error
    let config = get_config()?;

    let app = RelmApp::new("net.uku3lig.Tyrolienne");
    app.run_async::<Tyrolienne>(config);

    Ok(())
}

fn get_config() -> Result<Config> {
    let Some(config_dir) = dirs::config_dir() else {
        bail!("could not get config dir!");
    };

    std::fs::create_dir_all(&config_dir)?;

    let config_file_path = config_dir.join("tyrolienne.toml");
    let config_contents = match std::fs::read_to_string(&config_file_path) {
        Ok(str) => str,
        Err(e) if e.kind() == std::io::ErrorKind::NotFound => {
            let default_cfg = toml::to_string(&Config::default())?;
            std::fs::write(&config_file_path, &default_cfg)?;
            bail!("config file was not found, wrote default one at {config_file_path:?}");
        }
        Err(e) => bail!(e),
    };

    let config: Config = toml::from_str(&config_contents)?;
    if config.zipline_token.is_empty() {
        bail!("zipline token is empty! please set it at {config_file_path:?}");
    }
    if config.zipline_url.is_empty() {
        bail!("zipline url is empty! please set it at {config_file_path:?}");
    }

    Ok(config)
}

async fn the_process(app: &Tyrolienne) -> Result<String> {
    if let Some(folder) = app.folder.as_ref() {
        tracing::info!("uploading to folder '{}'...", folder.name);
    } else {
        tracing::info!("uploading video...");
    }

    let Some(ref video_path) = app.video_path else {
        bail!("No video given!");
    };

    let res = zipline::upload_file(&app.config, app.folder.as_ref(), video_path).await?;
    let zp_file = &res.files[0];

    tracing::info!("recalculating thumbnails...");

    zipline::recalc_thumbnails(&app.config).await?;

    tokio::time::sleep(Duration::from_secs(2)).await;

    tracing::info!("fetching thumbnail url...");

    let res = zipline::get_file_details(&app.config, &zp_file.id).await?;
    let thumbnail_url = res
        .thumbnail_url(&app.config)
        .ok_or_eyre("could not get thumbnail url")?;

    // TODO get w&h from video
    Ok(format!(
        "https://autocompressor.net/av1?v={}&i={}&w=1920&h=1080",
        Encoded(&zp_file.url),
        Encoded(&thumbnail_url)
    ))
}
